"""Types used in {mod}`~boilercv.correlations`."""

from collections.abc import Callable, Mapping
from dataclasses import asdict, dataclass
from pathlib import Path
from types import SimpleNamespace
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    _LiteralGenericAlias,  # pyright: ignore[reportAttributeAccessIssue]
    overload,
)

import sympy
from numpydantic import NDArray, Shape
from pydantic import (
    AfterValidator,
    BeforeValidator,
    PlainValidator,
    SerializerFunctionWrapHandler,
    ValidationInfo,
    WrapSerializer,
    WrapValidator,
)
from sympy import sympify
from sympy.logic.boolalg import BooleanAtom

from boilercv.morphs.contexts import ContextValue

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# * MARK: Type aliases
Corr: TypeAlias = Literal["beta", "nusselt"]
"""Correlation type."""
Range: TypeAlias = Literal["range"]
"""Range type."""
LiteralKeys: TypeAlias = _LiteralGenericAlias
"""Keys."""
# ? Notebook handling
NbProcess: TypeAlias = Callable[[Path, SimpleNamespace], None]
"""Notebook process."""
Stage: TypeAlias = Literal["large_sources", "sources", "filled"]
"""Stage."""
# ? `numpy` array shapes
AnyShape: TypeAlias = Shape["*"]  # noqa: F722
"""Any shape."""
Vector: TypeAlias = Shape["*"]  # noqa: F722
"""Vector shape."""
# ? Equations
Expectation: TypeAlias = float | NDArray[Vector, float]  # pyright: ignore[reportInvalidTypeArguments]
"""Expectation."""

# * MARK: Type variables

P = TypeVar("P", contravariant=True)
"""Contravariant type to represent parameters."""
R = TypeVar("R", covariant=True)
"""Covariant type to represent returns."""
Ps = ParamSpec("Ps")
"""Parameter type specification."""
K = TypeVar("K")
"""Key type."""
V = TypeVar("V")
"""Value type."""
CV = TypeVar("CV", bound=ContextValue, contravariant=True)
"""Context value type."""
CVL = TypeVar("CVL", bound="DataclassInstance | Mapping[Any, Any]", contravariant=True)
"""Context value-like type."""

# * MARK: Protocols


class Transform(Protocol[P, Ps, R]):  # noqa: D101
    def __call__(self, i: P, /, *args: Ps.args, **kwds: Ps.kwargs) -> R: ...  # noqa: D102


# * MARK: Validator


def validator(context_value_type: type[CV]):
    """Transform function expecting `context_value_type` to a Pydantic validator form.

    Get {class}`~boilercv.morphs.types.runtime.ContextValue` of `context_value_type` from `ValidationInfo` and pass to function expecting `context_value_type`.
    """

    def validator_maker(f: Callable[[P, CV], R]) -> Callable[[P, ValidationInfo], R]:
        def validate(v: P, info: ValidationInfo, /) -> R:
            key = context_value_type.name_to_snake()
            context = info.context or {}
            if not context:
                raise ValueError(
                    f"No context given. Expected value at '{key}' of type '{context_value_type}'."
                )
            context_value = context.get(key)
            if not context_value:
                raise ValueError(
                    f"No context value at {key}. Expected context value at '{key}' of type '{context_value_type}'."
                )
            if not isinstance(context_value, context_value_type):
                raise ValueError(  # noqa: TRY004 so Pydantic catches it
                    f"Context value at {key} not of expected type '{context_value_type}'."
                )
            return f(v, context_value)

        return validate

    return validator_maker


Kind: TypeAlias = Literal["latex", "sympy"]
"""Kind."""
Sym: TypeAlias = Literal[
    "Nu_c", "Fo_0", "Ja", "Re_b", "Re_b0", "Pe", "Pr", "beta", "alpha", "pi"
]
"""Symbol."""
Param: TypeAlias = Literal[
    "nusselt",
    "bubble_fourier",
    "bubble_jakob",
    "bubble_reynolds",
    "bubble_initial_reynolds",
    "bubble_peclet",
    "liquid_prandtl",
    "beta",
    "thermal_diffusivity",
    "pi",
]
"""Parameter."""
Equation: TypeAlias = Literal[
    "florschuetz_chao_1965",
    "isenberg_sideman_1970",
    "akiyama_1973",
    "chen_mayinger_1992",
    "zeitoun_et_al_1995",
    "kalman_mori_2002",
    "warrier_et_al_2002",
    "yuan_et_al_2009",
    "lucic_mayinger_2010",
    "kim_park_2011",
    "al_issa_et_al_2014",
    "tang_et_al_2016",
]
"""Equation."""
ExtraEquation: TypeAlias = Literal[
    "ranz_marshall_1952",  # Requires numerical solution
    "florschuetz_chao_1965_2",  # Requires numerical solution
    "hughmark_1967",  # Requires numerical solution
    "hughmark_1967_2",  # Requires numerical solution
    "inaba_et_al_2013",  # Uses custom `epsilon` and void fraction
]
"""Extended equation."""
AnyEquation: TypeAlias = Equation | ExtraEquation
"""Any equation."""

# * MARK: Serializers

WhenUsed: TypeAlias = Literal["always", "unless-none", "json", "json-unless-none"]


def StrSerializer(when_used: WhenUsed = "always") -> WrapSerializer:  # noqa: N802; Can't inherit from frozen
    """Serialize as string."""
    return WrapSerializer(_str, when_used=when_used)


def _str(v: Any, nxt: SerializerFunctionWrapHandler):
    """Stringify or try next serializer."""
    try:
        return str(v)
    except ValueError:
        return nxt(v)


# * MARK: Validators


Validator: TypeAlias = Literal["before", "wrap", "after", "plain"]
"""Validators."""


@overload
def TypeValidator(typ: type[K], mode: Literal["before"]) -> BeforeValidator: ...
@overload
def TypeValidator(typ: type[K], mode: Literal["wrap"]) -> WrapValidator: ...
@overload
def TypeValidator(typ: type[K], mode: Literal["after"]) -> AfterValidator: ...
@overload
def TypeValidator(typ: type[K], mode: Literal["plain"]) -> PlainValidator: ...
@overload
def TypeValidator(typ: type[K]) -> PlainValidator: ...
def TypeValidator(typ: type[K], mode: Validator = "plain") -> Any:  # noqa: N802; Can't inherit from frozen
    """Validate type."""

    def validate(v: K) -> K:
        if isinstance(v, typ):
            return v
        raise ValueError(f"Input should be a valid {typ}")

    return VALIDATORS[mode](validate)


VALIDATORS: dict[Validator, Any] = {
    "before": BeforeValidator,
    "wrap": WrapValidator,
    "after": AfterValidator,
    "plain": PlainValidator,
}
"""Validators."""

S = TypeVar(
    "S", bound=Annotated[LiteralKeys, TypeValidator(str), StrSerializer("json")] | str
)
"""Keys representing symbols."""


@dataclass
class SympifyParams(ContextValue):
    """Sympify parameters."""

    locals: dict[str, Any] | None = None
    convert_xor: bool | None = None
    strict: bool = False
    rational: bool = False
    evaluate: bool | None = None


def sympify_expr(expr: str | sympy.Eq, sympify_params: SympifyParams):
    """Sympify an expression from local variables."""
    return sympify(expr, **asdict(sympify_params))


# * MARK: Annotated types for Pydantic


Basic: TypeAlias = Annotated[
    sympy.Basic, TypeValidator(sympy.Basic), StrSerializer("json")
]
"""Annotated {class}`~sympy.core.basic.Basic` suitable for use in Pydantic models."""


trivial = sympy.Eq(1, 0, evaluate=False)
"""Trivial equation that won't evaluate to a `True` instance of `sympy.boolean.linalg.BooleanAtom`."""

Eq: TypeAlias = Annotated[
    Basic,
    BeforeValidator(validator(SympifyParams)(sympify_expr)),
    BeforeValidator(
        lambda v: (
            sympy.Eq(1, 0, evaluate=False)
            if isinstance(v, BooleanAtom) or v == trivial or not v
            else v
        )
    ),
    TypeValidator(sympy.Eq, "after"),
    # ? SymPy equations sometimes evaluate to `bool` on copy
    WrapSerializer(
        lambda v, nxt: "" if isinstance(v, BooleanAtom) or v == trivial else nxt(v),
        when_used="json",
    ),
]
"""{data}`~boilercv.correlations.types.Basic` narrowed to {class}`~sympy.core.relational.Eq`."""

Expr: TypeAlias = Annotated[
    Basic,
    BeforeValidator(validator(SympifyParams)(sympify_expr)),
    TypeValidator(sympy.Expr, "after"),
]
"""{data}`~boilercv.correlations.types.Basic` narrowed to {class}`~sympy.core.relational.Eq`."""

trivial_expr = sympy.Expr()
"""Trivial boolean that won't evaluate to a `True` instance of `sympy.boolean.linalg.BooleanAtom`."""


AnyExpr: TypeAlias = Annotated[
    Basic,
    BeforeValidator(validator(SympifyParams)(sympify_expr)),
    BeforeValidator(
        lambda v: (
            sympy.Expr()
            if isinstance(v, BooleanAtom) or v == trivial_expr or not v
            else v
        )
    ),
    # ? SymPy equations sometimes evaluate to `bool` on copy
    WrapSerializer(
        lambda v, nxt: ""
        if isinstance(v, BooleanAtom) or v == trivial_expr
        else nxt(v),
        when_used="json",
    ),
]
"""{data}`~boilercv.correlations.types.Basic` narrowed to {class}`~sympy.core.relational.Eq`."""

EQ = TypeVar("EQ", bound=Eq | str)
"""SymPy symbolic equation or string type."""



"""Theoretical correlations for bubble lifetimes."""

from collections.abc import Callable
from pathlib import Path
from tomllib import loads
from typing import Any, cast, get_args

import numpy
import sympy
from sympy.logic.boolalg import Boolean

from boilercv.correlations.models import (
    Correlation,
    Equations,
    SolvedEquations,
    SymbolicCorrelation,
)
from boilercv.correlations.types import AnyExpr, Equation, Sym, trivial_expr
from boilercv.morphs.types import LiteralGenericAlias

_base = Path(__file__).with_suffix(".toml")
META_TOML = _base.with_stem("meta")
RANGES_TOML = _base.with_stem("ranges")
SYMBOLS = {
    "Nu_c": "nusselt",
    "Fo_0": "bubble_fourier",
    "Ja": "bubble_jakob",
    "Re_b": "bubble_reynolds",
    "Re_b0": "bubble_initial_reynolds",
    "Pr": "liquid_prandtl",
    "beta": "beta",
    "Pe": "peclet",
    "alpha": "thermal_diffusivity",
    "pi": "pi",
}

GROUPS = {
    k: f"Group {v}"
    for k, v in {
        "florschuetz_chao_1965": 1,
        "isenberg_sideman_1970": 3,
        "akiyama_1973": 3,
        "chen_mayinger_1992": 3,
        "zeitoun_et_al_1995": 5,
        "kalman_mori_2002": 3,
        "warrier_et_al_2002": 4,
        "yuan_et_al_2009": 4,
        "lucic_mayinger_2010": 3,
        "kim_park_2011": 3,
        # "inaba_et_al_2013": 5,
        "al_issa_et_al_2014": 3,
        "tang_et_al_2016": 3,
    }.items()
}


def get_equations(path: Path) -> Equations[AnyExpr]:
    """Get equations."""
    return Equations[AnyExpr].context_model_validate(
        obj=loads(path.read_text("utf-8") if path.exists() else ""),
        context=Equations[AnyExpr].get_context(symbols=get_args(Sym)),
    )


def get_equations_and_solutions(
    equations: Path, solutions: Path, solve_sym: LiteralGenericAlias
) -> dict[Equation, SymbolicCorrelation]:
    """Get correlation equations and their solutions."""
    context = SolvedEquations[str].get_context(
        symbols=get_args(Sym), solve_syms=get_args(solve_sym)
    )
    ranges = get_equations(RANGES_TOML)
    return {
        name: SymbolicCorrelation(
            expr=cast(sympy.Expr, next(iter(soln.values())).solutions[0]),
            range=cast(
                Boolean,
                None if ranges[name].sympy == trivial_expr else ranges[name].sympy,
            ),
        )
        for name, soln in (
            SolvedEquations[solve_sym]
            .context_model_validate(
                dict(
                    equations=loads(
                        equations.read_text("utf-8") if equations.exists() else ""
                    ),
                    solutions=loads(
                        solutions.read_text("utf-8") if solutions.exists() else ""
                    ),
                ),
                context=context,
            )
            .solutions
        ).items()
    }


def get_correlations(
    equations: Path, solutions: Path, solve_sym: LiteralGenericAlias
) -> dict[Equation, Correlation]:
    """Get correlations and their ranges."""
    return {
        name: Correlation(
            expr=lambdify_expr(soln.expr),
            range=lambdify_expr(soln.range) if soln.range else None,
        )
        for name, soln in get_equations_and_solutions(
            equations=equations, solutions=solutions, solve_sym=solve_sym
        ).items()
    }


def lambdify_expr(expr: sympy.Basic) -> Callable[..., Any]:
    """Get symbolic functions."""
    return sympy.lambdify(
        expr=expr,
        modules=numpy,
        args=[s for s in expr.free_symbols if s.name in SYMBOLS],  # type: ignore
    )



"""Types relevant to array manipulation and image processing."""

from typing import Any, Protocol, TypeAlias, TypeVar

from numpy import (
    bool_,
    datetime64,
    floating,
    generic,
    integer,
    number,
    timedelta64,
    typing,
)
from numpy.typing import NBitBase, NDArray
from pandas import DataFrame
from xarray import DataArray, Dataset

DA: TypeAlias = DataArray
DF: TypeAlias = DataFrame
DS: TypeAlias = Dataset

ArrayLike: TypeAlias = typing.ArrayLike
"""Array-like."""

Arr: TypeAlias = NDArray[generic]
"""Generic array type. Consistent with OpenCV's type annotations."""
ArrBool: TypeAlias = NDArray[bool_]
"""A boolean array."""
ArrFloat: TypeAlias = NDArray[floating[NBitBase]]
"""An integer array with arbitrary bit depth."""
ArrInt: TypeAlias = NDArray[integer[NBitBase]]
"""An integer array."""
ArrNum: TypeAlias = NDArray[number[NBitBase]]
"""A number array."""
ArrDT: TypeAlias = NDArray[datetime64]
"""Datetime array type."""
ArrTD: TypeAlias = NDArray[timedelta64]
"""Timedelta array type."""

Img: TypeAlias = ArrInt
"""An integer array representing an image."""
ImgBool: TypeAlias = ArrBool
"""A boolean array representing an image mask."""
ImgLike: TypeAlias = ArrayLike
"""An array-like object representable as an image."""

Vid: TypeAlias = Img
"""An integer array representing a video."""
VidBool: TypeAlias = ImgBool
"""A boolean array representing a video mask."""


class SupportsMul(Protocol):
    """Protocol for types that support multiplication."""

    def __mul__(self, other: Any) -> Any: ...


DA_T = TypeVar("DA_T", bound=DA)



"""Tools for datasets."""

from collections.abc import Callable, Sequence
from itertools import chain
from typing import Any, TypedDict

from numpy import full
from pandas import IndexSlice
from pytz import timezone
from xarray import DataArray, Dataset, apply_ufunc

from boilercv.data.models import Dimension, get_dims
from boilercv.types import DA, DS, ArrayLike

VIDEO = "video"
"""Name of the video array in a dataset."""
ROI = "roi"
"""Name of the ROI array in a dataset."""
HEADER = "header"
"""Name of the header metadata (attache to an empty array) in a dataset."""
TIMEZONE = timezone("US/Pacific")
"""Timezone for all data."""

FRAME = "frame"
"""Frame dimension name."""
TIME = "time"
"""Time dimension name."""
UTC_TIME = "utc"
"""UTC time dimension name."""

Y = "y"
"""Y dimension name."""
X = "x"
"""X dimension name."""
YX = ["y", "x"]
"""Lenth dimension names."""

PX = "px"
"""Pixel length dimension suffix."""
YPX = f"{Y}{PX}"
"""Y pixel length dimension name."""
XPX = f"{X}{PX}"
"""X pixel length dimension name."""
YX_PX = [YPX, XPX]
"""Pixel length dimension names."""

DIMS = [FRAME, YPX, XPX]
"""Initial dimensions in a dataset."""

PACKED_DIM_INDEX = 2
"""Index of the packed dimension in the dimensions list. Corresponds to X."""
PACKED = "packed"
"""Packed suffix."""
XPX_PACKED = f"xpx_{PACKED}"
"""X pixel length dimension name when bit-packed."""
PACKED_DIMS = [FRAME, YPX, XPX_PACKED]
"""Initial dimensions in a dataset when bit-packed."""

VIDEO_NAME = "video_name"
"""Dimension for the video name, for datasets with frames from multiple videos."""

LENGTH = "um"
"""Length dimension units."""
SAMPLE_DIAMETER_UM = 9_525_000
"""Sample diameter in micrometers."""

islice = IndexSlice
"""Helper for slicing multi-index dataframes."""


def identity_da(da: DA, dim: str) -> DA:
    """Construct a data array that maps a dimension's coordinates to itself.

    Useful to apply `xarray.apply_ufunc` along coordinate values.

    Args:
        da: Data array.
        dim: The dimension to extract.
    """
    return DataArray(dims=(dim), coords={dim: da[dim].values}, data=da[dim])


class CommonKwargs(TypedDict):
    """Keyword arguments common to `xarray.apply_ufunc` calls."""

    input_core_dims: list[list[str]]
    """Dimensions which will remain after the function is applied."""
    vectorize: bool
    """Whether to vectorize the function over the input dimensions."""
    keep_attrs: bool
    """Whether to keep attributes from the input data array."""
    kwargs: dict[str, Any] | None
    """Keyword arguments for the function to be wrapped."""


def apply_to_img_da(
    func: Callable[..., Any],
    *args: Any,
    returns: int | None = 1,
    vectorize: bool = False,
    name: Sequence[str] | str = "",
    kwargs: dict[str, Any] | None = None,
) -> Any:
    """Apply functions that transform images to transform data arrays instead."""
    core_dims = [YX_PX]
    common_kwargs = CommonKwargs(
        input_core_dims=core_dims * len(args),
        vectorize=vectorize,
        keep_attrs=True,
        kwargs=kwargs,
    )
    if returns and returns > 0:
        result = apply_ufunc(
            func, *args, **common_kwargs, output_core_dims=core_dims * returns
        )
        if name and returns == 1:
            result = result.rename(name)
        if name and returns > 1:
            result = tuple(
                part.rename(name_) for name_, part in zip(name, result, strict=True)
            )
        return result
    if not returns or returns == 0:
        apply_ufunc(func, *args, **common_kwargs)


def assign_ds(
    name: str,
    data: ArrayLike,
    dims: Sequence[Dimension],
    secondary_dims: Sequence[Dimension] = (),
    fixed_dims: Sequence[Dimension] = (),
    fixed_secondary_dims: Sequence[Dimension] = (),
    ds: DS | None = None,
    long_name: str = "",
    units: str = "",
) -> DS:
    """Build a data array and assign it to a dataset."""
    if not ds:
        ds = Dataset()
    da = build_da(
        name,
        data,
        dims,
        secondary_dims,
        fixed_dims,
        fixed_secondary_dims,
        long_name,
        units,
    )
    ds[name] = da
    return ds


def build_da(
    name: str,
    data: ArrayLike,
    dims: Sequence[Dimension],
    secondary_dims: Sequence[Dimension] = (),
    fixed_dims: Sequence[Dimension] = (),
    fixed_secondary_dims: Sequence[Dimension] = (),
    long_name: str = "",
    units: str = "",
):
    """Build a data array."""
    all_primary_dims = chain(fixed_dims, dims)
    all_fixed_dims = chain(fixed_dims, fixed_secondary_dims)
    all_variable_dims = chain(dims, secondary_dims)
    if not long_name:
        long_name = name.capitalize() if len(name) > 1 else name
    if isinstance(data, Sequence) and not len(data):
        shape: list[int] = []
        for dim in all_primary_dims:
            if dim.coords is None:
                shape.append(1)
            else:
                shape.append(len(dim.coords))
        data = full(shape, None)
    attrs = {"long_name": long_name}
    if units:
        attrs["units"] = units
    da = DataArray(name=name, dims=get_dims(*all_primary_dims), data=data, attrs=attrs)
    for dim in chain(all_fixed_dims, all_variable_dims):
        da = dim.assign_to(da)
    return da



"""Morph."""

from __future__ import annotations

from collections.abc import Iterator, Mapping, MutableMapping
from contextlib import suppress
from typing import (
    Any,
    ClassVar,
    Generic,
    Self,
    TypeVar,
    get_args,
    get_type_hints,
    overload,
)

from pydantic import ConfigDict, Field, RootModel, ValidationError, model_validator

from boilercv.morphs.types import (
    RK,
    RV,
    DictDict,
    DictType,
    HasModelDump,
    K,
    MapMap,
    MapType,
    Ps,
    R,
    TypeDict,
    TypeMap,
    Types,
    TypeType,
    V,
)

TRUNC = 200
"""Truncate representations beyond this length."""


class Morph(RootModel[MutableMapping[K, V]], MutableMapping[K, V], Generic[K, V]):  # noqa: PLR0904
    """Type-checked, generic, morphable mapping."""

    root: MutableMapping[K, V] = Field(default_factory=dict)
    """Type-checked dictionary as the root data."""
    model_config: ClassVar = ConfigDict(protected_namespaces=("model_", "morph_"))
    """Root configuration, merged with subclass configs."""

    def __repr__(self):
        root = str(self.root)
        return (
            f"{type(self).__name__}(" + root[:TRUNC] + "...)"
            if len(root) > TRUNC
            else f"{type(self).__name__}({self.root})"
        )

    @classmethod
    def morph_get_inner_types(cls) -> Types:
        """Get types of the keys and values."""
        return Types(*get_args(cls.model_fields["root"].annotation))

    @model_validator(mode="before")
    @classmethod
    def morph_check_parametrized(cls, morph):
        """Qualify."""
        if cls.__pydantic_generic_metadata__["parameters"]:
            raise ValueError("Cannot instantiate generic morphs.")
        return morph

    # ! (([K, V] -> [K, V]) -> Self)
    @overload
    def morph_cpipe(
        self,
        f: MapMap[K, V, K, V, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    @overload
    def morph_cpipe(
        self,
        f: DictDict[K, V, K, V, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    # ! ((Self -> [K, V]) -> Self)
    @overload
    def morph_cpipe(
        self,
        f: TypeMap[Self, K, V, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    @overload
    def morph_cpipe(
        self,
        f: TypeDict[Self, K, V, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    # ! ((Self -> [RK, RV]) -> Morph[RK, RV])
    @overload
    def morph_cpipe(
        self,
        f: TypeMap[Self, RK, RV, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Morph[RK, RV]: ...
    @overload
    def morph_cpipe(
        self,
        f: TypeDict[Self, RK, RV, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Morph[RK, RV]: ...
    # ! (([K, V] -> R) -> R)
    @overload
    def morph_cpipe(
        self,
        f: MapType[K, V, R, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> R: ...
    @overload
    def morph_cpipe(
        self,
        f: DictType[K, V, R, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> R: ...
    # ! ((Self -> Self) -> Self)
    @overload
    def morph_cpipe(
        self,
        f: TypeType[Self, Self, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    # ! ((Self -> R) -> R)
    @overload
    def morph_cpipe(
        self,
        f: TypeType[Self, R, Ps],
        morph_context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> R: ...
    # ! ((Any -> Any) -> Any)
    def morph_cpipe(
        self,
        f: TypeType[Any, Any, Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self | Morph[Any, Any] | Any:
        """Pipe with context."""
        context = dict(context)
        self_k, self_v = self.morph_get_inner_types()
        out_k, out_v = (Any, Any)
        if (type_hints := get_type_hints(f)) and (hint := type_hints.get("return")):
            hints = get_args(hint)
            if hints and len(hints) == 2:
                out_k, out_v = Types(*hints)
            with suppress(TypeError):
                if issubclass(hint, Morph):
                    out_k, out_v = hint.morph_get_inner_types()
        copy = self.model_validate(obj=self.model_dump(), context=context)
        result = f(copy, *args, **kwds)
        result = (
            result.model_dump(warnings="none")
            if isinstance(result, HasModelDump)
            else result
        )
        if not isinstance(result, Mapping) or not result:
            return result
        result = dict(result)
        k = get_morph_hint(self_k, out_k) or Any
        v = get_morph_hint(self_v, out_v) or Any
        if Types(k, v) == self.morph_get_inner_types():
            return self.model_validate(obj=result, context=context)
        with suppress(ValidationError, TypeError, ValueError):
            return Morph[k, v].model_validate(obj=result, context=context)
        return result

    # ! (([K] -> [K]) -> Self)
    @overload
    def morph_ckpipe(
        self,
        f: TypeType[list[K], list[K], Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    # ! (([K] -> [RK]) -> Morph[RK, V])
    @overload
    def morph_ckpipe(
        self,
        f: TypeType[list[K], list[RK], Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Morph[RK, V]: ...
    # ! (([K] -> [Any]) -> Self | Morph[Any, V]
    def morph_ckpipe(
        self,
        f: TypeType[list[K], list[Any], Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self | Morph[Any, V]:
        """Pipe, morphing keys."""

        def pipe(_, *args: Ps.args, **kwds: Ps.kwargs):
            return dict(
                zip(f(list(self.keys()), *args, **kwds), self.values(), strict=False)
            )

        return self.morph_cpipe(pipe, context, *args, **kwds)

    # ! (([V] -> [V]) -> Self)
    @overload
    def morph_cvpipe(
        self,
        f: TypeType[list[V], list[V], Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self: ...
    # ! (([V] -> [RV]) -> Morph[K, RV])
    @overload
    def morph_cvpipe(
        self,
        f: TypeType[list[V], list[RV], Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Morph[K, RV]: ...
    # ! (([V] -> [Any]) -> Self | Morph[K, Any]
    def morph_cvpipe(
        self,
        f: TypeType[list[V], list[Any], Ps],
        context: Mapping[str, Any],
        /,
        *args: Ps.args,
        **kwds: Ps.kwargs,
    ) -> Self | Morph[K, Any]:
        """Pipe, morphing values."""

        def pipe(_, *args: Ps.args, **kwds: Ps.kwargs):
            return dict(
                zip(self.keys(), f(list(self.values()), *args, **kwds), strict=False)
            )

        return self.morph_cpipe(pipe, context, *args, **kwds)

    # ! (([K, V] -> [K, V]) -> Self)
    @overload
    def morph_pipe(
        self, f: MapMap[K, V, K, V, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    @overload
    def morph_pipe(
        self, f: DictDict[K, V, K, V, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    # ! ((Self -> [K, V]) -> Self)
    @overload
    def morph_pipe(
        self, f: TypeMap[Self, K, V, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    @overload
    def morph_pipe(
        self, f: TypeDict[Self, K, V, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    # ! ((Self -> [RK, RV]) -> Morph[RK, RV])
    @overload
    def morph_pipe(
        self, f: TypeMap[Self, RK, RV, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Morph[RK, RV]: ...
    @overload
    def morph_pipe(
        self, f: TypeDict[Self, RK, RV, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Morph[RK, RV]: ...
    # ! (([K, V] -> R) -> R)
    @overload
    def morph_pipe(
        self, f: MapType[K, V, R, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> R: ...
    @overload
    def morph_pipe(
        self, f: DictType[K, V, R, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> R: ...
    # ! ((Self -> Self) -> Self)
    @overload
    def morph_pipe(
        self, f: TypeType[Self, Self, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    # ! ((Self -> R) -> R)
    @overload
    def morph_pipe(
        self, f: TypeType[Self, R, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> R: ...
    # ! ((Any -> Any) -> Any)
    def morph_pipe(
        self, f: TypeType[Any, Any, Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self | Morph[Any, Any] | Any:
        """Pipe."""
        return self.morph_cpipe(f, {}, *args, **kwds)

    # ! (([K] -> [K]) -> Self)
    @overload
    def morph_kpipe(
        self, f: TypeType[list[K], list[K], Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    # ! (([K] -> [RK]) -> Morph[RK, V])
    @overload
    def morph_kpipe(
        self, f: TypeType[list[K], list[RK], Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Morph[RK, V]: ...
    # ! (([K] -> [Any]) -> Self | Morph[Any, V]
    def morph_kpipe(
        self, f: TypeType[list[K], list[Any], Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self | Morph[Any, V]:
        """Pipe, morphing keys."""
        return self.morph_ckpipe(f, {}, *args, **kwds)

    # ! (([V] -> [V]) -> Self)
    @overload
    def morph_vpipe(
        self, f: TypeType[list[V], list[V], Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self: ...
    # ! (([V] -> [RV]) -> Morph[K, RV])
    @overload
    def morph_vpipe(
        self, f: TypeType[list[V], list[RV], Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Morph[K, RV]: ...
    # ! (([V] -> [Any]) -> Self | Morph[K, Any]
    def morph_vpipe(
        self, f: TypeType[list[V], list[Any], Ps], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> Self | Morph[K, Any]:
        """Pipe, morphing values."""
        return self.morph_cvpipe(f, {}, *args, **kwds)

    # `MutableMapping` methods adapted from `collections.UserDict`, but with `data`
    # replaced by `root`and `hasattr` guard changed to equivalent `getattr(..., None)`
    # pattern in `__getitem__`. This is done to prevent inheriting directly from
    # `UserDict`, which doesn't play nicely with `pydantic.RootModel`.
    # Source: https://github.com/python/cpython/blob/7d7eec595a47a5cd67ab420164f0059eb8b9aa28/Lib/collections/__init__.py#L1121-L1211

    @classmethod
    def fromkeys(cls, iterable, value=None):  # noqa: D102
        return cls(dict.fromkeys(iterable, value))

    def __len__(self):
        return len(self.root)

    def __getitem__(self, key: K) -> V:
        if key in self.root:
            return self.root[key]
        if missing := getattr(self.__class__, "__missing__", None):
            return missing(self, key)
        raise KeyError(key)

    def __iter__(self) -> Iterator[K]:  # pyright: ignore[reportIncompatibleMethodOverride] iterate over `root` instead of `self`
        return iter(self.root)

    def __setitem__(self, key: K, item: V):
        self.root[key] = item

    def __delitem__(self, key: K):
        del self.root[key]

    # Modify __contains__ to work correctly when __missing__ is present
    def __contains__(self, key: K):  # pyright: ignore[reportIncompatibleMethodOverride]
        return key in self.root

    def __or__(self, other: HasModelDump | Mapping[Any, Any] | Any) -> Self:
        if isinstance(other, Mapping) and isinstance(other, HasModelDump):
            return self.model_construct(self.model_dump() | other.model_dump())
        if isinstance(other, Mapping):
            return self.model_construct(self.model_dump() | other)
        return NotImplemented

    def __ror__(self, other: HasModelDump | Mapping[Any, Any] | Any) -> Self:
        if isinstance(other, Mapping) and isinstance(other, HasModelDump):
            return self.model_construct(other.model_dump() | self.model_dump())
        if isinstance(other, Mapping):
            return self.model_construct(other | self.model_dump())
        return NotImplemented

    def __ior__(self, other) -> Self:
        return self | other


def get_morph_hint(
    in_hint: type, out_hint: type | TypeVar | None = None
) -> type | None:
    """Handle missing and {class}`~typing.TypeVar` hints."""
    if in_hint is out_hint:
        return in_hint
    if not out_hint or isinstance(out_hint, TypeVar):
        return in_hint
    return out_hint



"""Process images with OpenCV."""

from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum

from cv2 import (
    ADAPTIVE_THRESH_MEAN_C,
    BORDER_CONSTANT,
    CHAIN_APPROX_NONE,
    FILLED,
    FLOODFILL_MASK_ONLY,
    MORPH_BLACKHAT,
    MORPH_CLOSE,
    MORPH_CROSS,
    MORPH_DILATE,
    MORPH_ELLIPSE,
    MORPH_ERODE,
    MORPH_GRADIENT,
    MORPH_HITMISS,
    MORPH_OPEN,
    MORPH_RECT,
    MORPH_TOPHAT,
    RETR_EXTERNAL,
    THRESH_BINARY,
    LineSegmentDetector,
    adaptiveThreshold,
    add,
    bitwise_not,
    copyMakeBorder,
    createLineSegmentDetector,
    cvtColor,
    drawContours,
    fillPoly,
    findContours,
    floodFill,
    getStructuringElement,
    morphologyEx,
)
from numpy import array, flip, fliplr, iinfo, zeros_like

from boilercv.colors import WHITE, WHITE3
from boilercv.images import unpad
from boilercv.types import ArrFloat, ArrInt, Img, ImgBool


def convert_image(img: Img, code: int | None = None) -> Img:
    """Convert image format, handling inconsistent type annotations."""
    return cvtColor(img, code)  # type: ignore  # pyright 1.1.333


def apply_mask(img: Img, mask: Img) -> Img:
    """Mask an image, keeping parts where the mask is bright."""
    return add(img, bitwise_not(mask))  # type: ignore  # pyright 1.1.333


def pad(img: Img, pad_width: int, value: int) -> Img:
    """Pad an image with a constant value.

    Parameters
    ----------
    img
        Image.
    pad_width
        Width of the pad, in pixels.
    value
        Pixel value to fill the pad with.

    Returns
    -------
    img:
        Padded image.
    """
    return copyMakeBorder(
        img,
        top=pad_width,
        bottom=pad_width,
        left=pad_width,
        right=pad_width,
        borderType=BORDER_CONSTANT,
        value=value,  # type: ignore  # pyright 1.1.333
    )


def binarize(img: Img, block_size: int = 11, thresh_dist_from_mean: int = 2) -> ImgBool:
    """Binarize an image with an adaptive threshold."""
    block_size += 1 if block_size % 2 == 0 else 0
    return adaptiveThreshold(
        src=img,
        maxValue=iinfo(img.dtype).max,
        adaptiveMethod=ADAPTIVE_THRESH_MEAN_C,
        thresholdType=THRESH_BINARY,
        blockSize=block_size,
        C=thresh_dist_from_mean,
    ).astype(bool)


def flood(img: Img) -> ImgBool:
    """Flood the image, returning the resulting flood as a bright mask."""
    seed_point = array(img.shape) // 2
    _max_value = iinfo(img.dtype).max
    # OpenCV needs a masked array with a one-pixel pad
    pad_width = 1
    mask = pad(zeros_like(img), pad_width, value=1)
    _retval, _image, mask, _rect = floodFill(
        image=img,
        mask=mask,
        seedPoint=tuple(flip(seed_point)),  # OpenCV expects (x, y)
        newVal=None,  # type: ignore  # pyright 1.1.333
        flags=FLOODFILL_MASK_ONLY,
    )
    # Return the mask in original dimensions
    return unpad(mask, pad_width).astype(bool)


def close_and_erode(img: Img) -> Img:
    """Close holes, then erode."""
    return transform(img, [Transform(Op.close, 4), Transform(Op.erode, 9)]).astype(bool)


def get_wall(roi: Img) -> Img:
    """Dilate the ROI to get the wall."""
    return transform(roi, Transform(Op.dilate, 9)).astype(bool)


class Op(Enum):
    """A morphological transform operation."""

    black_hat = MORPH_BLACKHAT
    close = MORPH_CLOSE
    cross = MORPH_CROSS
    dilate = MORPH_DILATE
    ellipse = MORPH_ELLIPSE
    erode = MORPH_ERODE
    gradient = MORPH_GRADIENT
    hitmiss = MORPH_HITMISS
    open = MORPH_OPEN
    rect = MORPH_RECT
    top_hat = MORPH_TOPHAT


@dataclass
class Transform:
    """A morphological transform."""

    op: Op
    """The transform to perform."""
    size: int
    """The elliptical kernel size."""


def transform(img: Img, transforms: Transform | Sequence[Transform]) -> Img:
    """Apply morphological transforms to an image with a dark background."""
    transforms = [transforms] if isinstance(transforms, Transform) else transforms
    pad_width = max(transform.size for transform in transforms)
    # Explicitly pad out the image since cv2.morphologyEx boundary handling is weird
    img = pad(img, pad_width, value=0)
    for transform in transforms:
        img = morphologyEx(  # type: ignore  # pyright 1.1.333
            src=img,
            op=transform.op.value,
            kernel=getStructuringElement(MORPH_ELLIPSE, [transform.size] * 2),
        )
    return unpad(img, pad_width)


def build_mask_from_polygons(img: Img, contours: Sequence[ArrInt]) -> Img:
    """Build a mask from the intersection of a sequence of polygonal contours."""
    # OpenCV expects contours as shape (N, 1, 2) instead of (N, 2)
    contours = [fliplr(contour).reshape(-1, 1, 2) for contour in contours]
    blank = zeros_like(img)
    return fillPoly(  # type: ignore  # pyright 1.1.333
        img=blank,
        pts=contours,  # Expects a list of coordinates, we have just one
        color=WHITE3,
    )


def find_contours(img: Img, method: int = CHAIN_APPROX_NONE) -> list[ArrInt]:
    """Find external contours of bright objects in an image."""
    contours, _hierarchy = findContours(
        image=img,
        mode=RETR_EXTERNAL,  # No hierarchy needed because we keep external contours
        method=method,
    )
    # Despite images having dims (y, x) and shape (h, w), OpenCV returns contours with
    # dims (point, 1, pair), where dim "pair" has coords (x, y).
    contours = [fliplr(contour.reshape(-1, 2)) for contour in contours]
    return contours  # type: ignore  # pyright 1.1.347


def draw_contours(
    img: Img,
    contours: Sequence[ArrInt],
    contour_index: int = -1,
    thickness: int = FILLED,
    color: int | tuple[int, ...] = WHITE,
) -> Img:
    """Draw contours on an image."""
    # OpenCV expects contours as shape (N, 1, 2) instead of (N, 2)
    contours = [fliplr(contour).reshape(-1, 1, 2) for contour in contours]
    return drawContours(
        image=img,
        contours=contours,
        contourIdx=contour_index,
        color=color,  # type: ignore  # pyright 1.1.333
        thickness=thickness,
    )


def find_line_segments(img: Img) -> tuple[ArrFloat, LineSegmentDetector]:
    """Find line segments in an image."""
    lsd = createLineSegmentDetector()
    lines, *_ = lsd.detect(img)
    # OpenCV returns line segments as shape (N, 1, 4) instead of (N, 4)
    lines = lines.reshape(-1, 4)
    return lines, lsd  # type: ignore  # pyright 1.1.333



"""Dataframes."""

from numpy import split
from pandas import DataFrame, concat

from boilercv.data import YX_PX
from boilercv.types import DF, ArrayLike


def df_points(points: ArrayLike, dims: list[str] = YX_PX) -> DF:
    """Build a dataframe from an array of points."""
    return (
        DataFrame(
            columns=dims,
            data=points,  # type: ignore  # pyright 1.1.333
        )
        .rename_axis(axis="index", mapper="point")
        .rename_axis(axis="columns", mapper="dim")
    )


def frame_lines(lines: ArrayLike) -> DF:
    """Build a dataframe from an array of line segments."""
    ordered_pairs = [df_points(point) for point in split(lines, 2, axis=1)]
    return (
        concat(axis="columns", keys=[0, 1], objs=ordered_pairs)
        .rename_axis(axis="index", mapper="line")
        .rename_axis(axis="columns", mapper=["coord", "dim"])
        .reorder_levels(axis="columns", order=["dim", "coord"])
    )



"""Types for static typing of mappings."""

from collections.abc import Mapping, MutableMapping
from typing import Any, TypeAlias, TypeVar

Leaf: TypeAlias = Any
"""Leaf node."""
Node: TypeAlias = Mapping[Any, "Node | Leaf"]
"""General node."""
MutableNode: TypeAlias = MutableMapping[Any, "MutableNode | Leaf"]
"""Mutable general node."""

T = TypeVar("T")
"""Type."""
K = TypeVar("K")
"""Key type."""
V = TypeVar("V")
"""Value type."""
SK = TypeVar("SK", bound=str)
"""String key type."""
MN = TypeVar("MN", bound=MutableNode)
"""Mutable node type."""



"""Mapping functions."""

from collections.abc import Callable, Iterable, Mapping, MutableMapping
from copy import deepcopy
from re import Pattern, sub
from typing import Any, Generic, NamedTuple

from boilercv.mappings.types import MN, SK, K, Leaf, Node, T, V


def apply(
    mapping: Mapping[Any, Any],
    node_fun: Callable[[Node], Any] = lambda n: n,
    leaf_fun: Callable[[Leaf], Any] = lambda v: v,
    node_cond: Callable[[Node], bool] = bool,
    leaf_cond: Callable[[Node], bool] = bool,
) -> dict[Any, Any]:
    """Apply functions and conditions recursively to nodes and leaves of a mapping."""
    filtered = dict(deepcopy(mapping))
    marks: list[Any] = []
    for k, v in filtered.items():
        if isinstance(v, Mapping):
            filtered[k] = apply(filtered[k], node_fun, leaf_fun)
            filtered[k] = node_fun(filtered[k])
            if not node_cond(filtered[k]):
                marks.append(k)
            continue
        filtered[k] = leaf_fun(filtered[k])
        if not leaf_cond(filtered[k]):
            marks.append(k)
    for mark in marks:
        del filtered[mark]
    return filtered


def filt(
    mapping: Mapping[Any, Any],
    node_cond: Callable[[Node], bool] = bool,
    leaf_cond: Callable[[Node], bool] = bool,
) -> dict[Any, Any]:
    """Filter nodes and leaves of a mapping recursively."""
    return apply(mapping, node_cond=node_cond, leaf_cond=leaf_cond)


def sort_by_keys_pattern(
    i: Mapping[SK, V],
    pattern: Pattern[str],
    groups: Iterable[str],
    apply_to_match: Callable[[list[str]], Any] = str,
    message: str = "Match not found when sorting.",
) -> dict[SK, V]:
    """Sort mapping by named grouping in keys pattern."""

    def get_key(item: tuple[str, Any]) -> str:
        if match := pattern.search(key := item[0]):
            return apply_to_match([match[g] for g in groups])
        raise ValueError(message.format(key=key))

    return dict(sorted(i.items(), key=get_key))


class Repl(NamedTuple, Generic[T]):
    """Contents of `dst` to replace with `src`, with `find` substrings replaced with `repl`."""

    src: T
    """Source identifier."""
    dst: T
    """Destination identifier."""
    find: str
    """Find this in the source."""
    repl: str
    """Replacement for what was found."""


def replace(i: dict[K, str], repls: Iterable[Repl[K]]) -> dict[K, str]:
    """Make replacements from `Repl`s."""
    for r in repls:
        i[r.dst] = i[r.src].replace(r.find, r.repl)
    return dict(i)


def replace_pattern(i: dict[K, str], repls: Iterable[Repl[K]]) -> dict[K, str]:
    """Make regex replacements."""
    for r in repls:
        i[r.dst] = sub(r.find, r.repl, i[r.src])
    return i


def sync(reference: Node | Leaf, target: MN | Leaf) -> MN:
    """Sync two mappings."""
    synced = deepcopy(target)
    for key in [k for k in synced if k not in reference]:
        del synced[key]
    for key in reference:
        if key in synced:
            if reference[key] == synced[key]:
                continue
            if isinstance(reference[key], dict) and isinstance(
                synced[key], MutableMapping
            ):
                synced[key] = sync(reference[key], synced[key])
                continue
        synced[key] = reference[key]
    return synced



"""Dimensionless parameters."""


def thermal_diffusivity(thermal_conductivity, density, isobaric_specific_heat):
    """Thermal diffusivity."""
    return thermal_conductivity / (density * isobaric_specific_heat)


def kinematic_viscosity(dynamic_viscosity, density):
    """Kinematic viscosity."""
    return dynamic_viscosity / density


def reynolds(velocity, characteristic_length, kinematic_viscosity):
    """Reynolds number."""
    return velocity * characteristic_length / kinematic_viscosity


def prandtl(dynamic_viscosity, isobaric_specific_heat, thermal_conductivity):
    """Prandtl number."""
    return (isobaric_specific_heat * dynamic_viscosity) / thermal_conductivity


def jakob(
    liquid_density,
    vapor_density,
    liquid_isobaric_specific_heat,
    subcooling,
    latent_heat_of_vaporization,
):
    """Jakob number."""
    return (liquid_density * liquid_isobaric_specific_heat * subcooling) / (
        vapor_density * latent_heat_of_vaporization
    )


def fourier(liquid_thermal_diffusivity, initial_bubble_diameter, time):
    """Fourier number."""
    return liquid_thermal_diffusivity * time / initial_bubble_diameter**2


def nusselt(heat_transfer_coefficient, characteristic_length, thermal_conductivity):
    """Nusselt number."""
    return heat_transfer_coefficient * characteristic_length / thermal_conductivity



"""Color constants for use in image processing and graphical user interfaces."""

# Standard RGB colors
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)

# OpenCV colors w/ color order BGR
RED_CV = (0, 0, 255)
GREEN_CV = (0, 255, 0)
BLUE_CV = (255, 0, 0)

# Colors
BLACK = 0
BLACK3 = (BLACK,) * 3
WHITE = 255
WHITE3 = (WHITE,) * 3



"""Nusselt number correlations for subcooled boiling bubble collapse."""

from pathlib import Path

from numpy import linspace, pi

from boilercv import correlations
from boilercv.correlations.models import Correlation, Expectations, SymbolicCorrelation
from boilercv.correlations.nusselt.types import SolveSym
from boilercv.correlations.types import Equation, Sym

PNGS = Path("data/png_equations_nusselt")
"""Equation PNGs."""

_base = Path(__file__).with_suffix(".toml")

EQUATIONS_TOML = _base.with_stem("equations")
"""TOML file with equations."""
EXPECTATIONS_TOML = _base.with_stem("expectations")
"""TOML file with expectations."""
SOLUTIONS_TOML = _base.with_stem("solutions")
"""TOML file with solutions."""
SYMBOL_EXPECTATIONS = Expectations[Sym].context_model_validate(
    obj={
        "Nu_c": 1.0,
        "Fo_0": linspace(start=0.0, stop=5.0e-3, num=10),
        "Ja": 1.0,
        "Re_b0": 100.0,
        "Pr": 1.0,
        "beta": 0.5,
        "alpha": 1.0,
        "pi": pi,
    },
    context=Expectations[Sym].get_context(),
)
"""Common keyword arguments applied to correlations."""


def get_equations_and_solutions() -> dict[Equation, SymbolicCorrelation]:
    """Get correlations."""
    return correlations.get_equations_and_solutions(
        EQUATIONS_TOML, SOLUTIONS_TOML, SolveSym
    )


def get_correlations() -> dict[Equation, Correlation]:
    """Get correlations."""
    return correlations.get_correlations(EQUATIONS_TOML, SOLUTIONS_TOML, SolveSym)



"""Contextual morphs."""

from __future__ import annotations

from collections import UserDict, defaultdict
from collections.abc import Callable, Mapping, Sequence
from contextlib import suppress
from dataclasses import dataclass, field
from functools import partial
from itertools import chain
from typing import Any, ClassVar, Generic, Self, get_args

from pydantic import BaseModel, ConfigDict, ValidationInfo, model_validator
from pydantic.alias_generators import to_snake

from boilercv.morphs.morphs import Morph
from boilercv.morphs.types import (
    CV,
    AnyPipe,
    ContextValueLike,
    HasModelDump,
    K,
    LiteralGenericAlias,
    Mode,
    Model,
    T,
    UnionGenericAlias,
    V,
)

# * MARK: Base context


class BaseContext:
    """Base class for contextualized models."""

    model_config: ClassVar = ConfigDict(
        protected_namespaces=(
            *(Morph.model_config.get("protected_namespaces") or []),
            "context_",
        )
    )

    @classmethod
    def context_model_validate(
        cls: type[Model],  # pyright: ignore[reportGeneralTypeIssues]
        obj: Any | None = None,
        *,
        strict: bool | None = None,
        from_attributes: bool | None = None,
        context: Context | None = None,
    ) -> Model:
        """Validate a pydantic model instance.

        Args:
            obj: The object to validate.
            strict: Whether to enforce types strictly.
            from_attributes: Whether to extract data from object attributes.
            context: Additional context to pass to the validator.

        Raises
        ------
            ValidationError: If the object could not be validated.

        Returns
        -------
            The validated model instance.
        """
        return cls.model_validate(  # pyright: ignore[reportAttributeAccessIssue]
            obj.model_dump() if isinstance(obj, HasModelDump) else obj or {},
            strict=strict,
            from_attributes=from_attributes,
            context=context or Context(),  # pyright: ignore[reportArgumentType]
        )

    @classmethod
    def context_model_validate_json(
        cls: type[Model],  # pyright: ignore[reportGeneralTypeIssues]
        json_data: str | bytes | bytearray = "{}",
        *,
        strict: bool | None = None,
        context: Context | None = None,
    ) -> Model:
        """Usage docs: https://docs.pydantic.dev/2.7/concepts/json/#json-parsing.

        Validate the given JSON data against the Pydantic model.

        Args:
            json_data: The JSON data to validate.
            strict: Whether to enforce types strictly.
            context: Extra variables to pass to the validator.

        Returns
        -------
            The validated Pydantic model.

        Raises
        ------
            ValueError: If `json_data` is not a JSON string.
        """
        return cls.model_validate_json(
            json_data,
            strict=strict,
            context=context or Context(),  # pyright: ignore[reportArgumentType]
        )


# * MARK: Contextualized base model


class ContextBaseModel(BaseModel, BaseContext):
    """Context base model."""

    @model_validator(mode="before")
    @classmethod
    def context_validate_before(
        cls, data: dict[str, Any], info: ValidationInfo
    ) -> dict[str, Any]:
        """Validate context before."""
        return cls.apply(mode="before", data=data, info=info)

    @model_validator(mode="after")
    def context_validate_after(self, info: ValidationInfo) -> Self:
        """Validate context after."""
        return self.apply(mode="after", data=self, info=info)

    @classmethod
    def apply(cls, mode: Mode | Any, data: T, info: ValidationInfo) -> T:
        """Apply a pipe to data."""
        context = Context(info.context) or Context()
        pipelines_context = (
            get_context_value(PipelineContext, context) or PipelineContext()
        )
        pipelines = pipelines_context.get(cls)  # pyright: ignore[reportArgumentType]
        if pipelines is None:
            return data
        match mode:
            case "after":
                pipeline = pipelines.after
            case "before":
                pipeline = pipelines.before
            case _:
                raise ValueError("Invalid mode.")
        if not pipeline:
            return data
        for pipe in pipeline:
            if isinstance(pipe, Pipe):
                data = pipe.f(data, pipe.context_value)
                continue
            if isinstance(pipe, PipeWithInfo):
                data = pipe.f(data, pipe.context_value, info)
                continue
            if isinstance(pipe, Callable):
                data = pipe(data)
                continue
            raise ValueError("Invalid pipe.")
        return data


# * MARK: Contextualized morph


class ContextMorph(Morph[K, V], BaseContext, Generic[K, V]):
    """Context morph."""

    @model_validator(mode="before")
    @classmethod
    def context_validate_before(
        cls, data: dict[K, V], info: ValidationInfo
    ) -> dict[K, V]:
        """Validate context before."""
        return cls.apply(mode="before", data=data, info=info)

    @model_validator(mode="after")
    def context_validate_after(self, info: ValidationInfo) -> Self:
        """Validate context after."""
        return self.apply(mode="after", data=self, info=info)

    @classmethod
    def apply(cls, mode: Mode, data: T, info: ValidationInfo) -> T:  # noqa: C901, PLR0912
        """Apply a pipe to data."""
        context = Context(info.context) or Context()
        pipelines_context = (
            get_context_value(PipelineContext, context) or PipelineContext()
        )
        pipelines = pipelines_context.get(cls)  # pyright: ignore[reportArgumentType]
        if pipelines is None:
            return data
        if mode == "before":
            pipeline = pipelines.before
            if not pipeline:
                return data
            for pipe in pipeline:
                if isinstance(pipe, Pipe):
                    data = pipe.f(data, pipe.context_value)
                    continue
                if isinstance(pipe, PipeWithInfo):
                    data = pipe.f(data, pipe.context_value, info)
                    continue
                if isinstance(pipe, Callable):
                    data = pipe(data)
                    continue
                raise ValueError("Invalid pipe.")
            return data
        pipeline = pipelines.after
        if not pipeline:
            return data
        for pipe in pipeline:
            if not pipeline:
                return data
            if isinstance(pipe, Pipe):
                data = data.morph_cpipe(pipe.f, context, pipe.context_value)  # pyright: ignore[reportAttributeAccessIssue]
                continue
            if isinstance(pipe, PipeWithInfo):
                data = data.morph_cpipe(pipe.f, context, pipe.context_value, info)  # pyright: ignore[reportAttributeAccessIssue]
                continue
            if isinstance(pipe, Callable):
                data = data.morph_cpipe(pipe, context)  # pyright: ignore[reportAttributeAccessIssue]
                continue
            raise ValueError("Invalid pipe.")
        return data

    @classmethod
    def compose_defaults(
        cls,
        mapping: dict[K, V | BaseModel] | None = None,
        keys: tuple[K, ...] | None = None,
        value: V | BaseModel | None = None,
        value_copier: Callable[[V | BaseModel], V | BaseModel] = lambda v: v,
        factory: Callable[..., Any] | None = None,
        value_model: type[ContextMorph[K, V] | ContextBaseModel | BaseModel]
        | None = None,
        value_context: Context | None = None,
    ) -> Context:
        """Compose defaults."""
        k, v = cls.morph_get_inner_types()
        if (
            not keys
            and isinstance(k, UnionGenericAlias)
            and all(isinstance(t, LiteralGenericAlias) for t in get_args(k))
        ):
            keys = tuple(chain.from_iterable([get_args(t) for t in get_args(k)]))
        if not keys and isinstance(k, LiteralGenericAlias):
            keys = get_args(k)
        with suppress(TypeError):
            if issubclass(v, BaseModel):
                value_model = value_model or v
        keys = keys or get_args(cls.morph_get_inner_types().key)
        value_context = value_context or Context()
        if value_model:
            factory = (
                partial(value_model.context_model_validate, context=value_context)
                if issubclass(value_model, ContextMorph | ContextBaseModel)
                else partial(
                    value_model.model_validate, obj={}, context=dict(value_context)
                )
            )
        defaults = (
            Defaults(mapping=mapping, value_copier=value_copier)
            if mapping
            else Defaults(
                keys=keys or (), value=value, factory=factory, value_copier=value_copier
            )
        )
        context = compose_context(
            PipelineContext({cls: Pipelines.make(before=Pipe(set_defaults, defaults))})
        )
        return compose_contexts(context, value_context)


# * MARK: Pipelines


@dataclass
class Pipelines:
    """Pipelines."""

    before: tuple[AnyPipe, ...] = field(default_factory=tuple)
    after: tuple[AnyPipe, ...] = field(default_factory=tuple)

    @classmethod
    def make(
        cls,
        before: Sequence[AnyPipe] | AnyPipe | None = None,
        after: Sequence[AnyPipe] | AnyPipe | None = None,
    ) -> Self:
        """Get validators."""
        before = () if before is None else before
        after = () if after is None else after
        bef = tuple(before) if isinstance(before, Sequence) else (before,)
        aft = tuple(after) if isinstance(after, Sequence) else (after,)
        return cls(before=bef, after=aft)


class ContextValue:
    """Context value."""

    @classmethod
    def name_to_snake(cls) -> str:
        """Get name."""
        return to_snake(cls.__name__)


class PipelineContext(UserDict[Any, Pipelines], ContextValue):
    """Morphs."""

    def __or__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: Self | Mapping[Any, Pipelines] | Any
    ) -> Self:
        if isinstance(other, PipelineContext):
            merged: dict[Any, Pipelines] = defaultdict(Pipelines)
            for model, pipeline in chain.from_iterable([self.items(), other.items()]):
                merged[model].before += pipeline.before
                merged[model].after += pipeline.after
            return type(self)(merged)
        if isinstance(other, Mapping):
            return self | type(self)(other)
        return NotImplemented

    def __ror__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: Self | Mapping[Any, Pipelines] | Any
    ) -> Self:
        if isinstance(other, Mapping):
            return type(self)(other) | self
        return NotImplemented

    def __ior__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: Self | Mapping[Any, Pipelines] | Any
    ) -> Self:
        self.data = (self | other).data
        return self


@dataclass
class Pipe:
    """Pipe."""

    f: Callable[[Any, Any], Any]
    context_value: ContextValue


@dataclass
class PipeWithInfo:
    """Pipe with validation info."""

    f: Callable[[Any, Any, Any], Any]
    context_value: ContextValue


# * MARK: Contexts


class Context(UserDict[str, ContextValueLike]):
    """Morphs."""

    @property
    def pipelines(self) -> PipelineContext:
        """Get pipelines."""
        context = self.get(get_context_key(PipelineContext), PipelineContext())
        return context if isinstance(context, PipelineContext) else PipelineContext()

    def __or__(self, other: Context | Mapping[str, Any] | Any) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        if isinstance(other, Context):
            merged = dict(self) | dict(other)
            if (morphs := get_context_value(PipelineContext, self)) is not None and (
                other_morphs := get_context_value(PipelineContext, other)
            ) is not None:
                merged[get_context_key(PipelineContext)] = morphs | other_morphs
            return type(self)(merged)
        if isinstance(other, Mapping):
            return self | type(self)(other)
        return NotImplemented

    def __ror__(self, other: Context | Mapping[str, Any] | Any) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        if isinstance(other, Mapping):
            return type(self)(other) | self
        return NotImplemented

    def __ior__(self, other: Context | Mapping[str, Any] | Any) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        self.data = (self | other).data
        return self


# * MARK: Context values


@dataclass
class Defaults(ContextValue, Generic[K, V]):
    """Context for expression validation."""

    mapping: dict[K, V] = field(default_factory=dict)
    """Mapping of keys to their default values."""
    keys: tuple[K, ...] = field(default_factory=tuple)
    """Default keys."""
    value: V | None = None
    """Default value."""
    factory: Callable[..., Any] | None = None
    """Default value factory."""
    value_copier: Callable[[V], V] | None = lambda v: v
    """Method to copy the default value."""


# * MARK: Context handlers


def set_defaults(i: dict[K, V], d: Defaults[K, V]) -> dict[K, V]:
    """Set `self.default_keys` using `self.default_factory` or `self.default`."""
    if not d.value_copier:
        d.value_copier = lambda v: v
    if d.mapping:
        if d.keys or d.value or d.factory:
            raise ValueError(
                "Cannot specify both 'mapping' and 'keys', 'value', or 'factory'."
            )
        return {key: d.value_copier(v) for key, v in d.mapping.items()} | i
    if not d.keys:
        raise ValueError("Must supply either 'mapping' or 'keys'.")
    if any(key not in i for key in d.keys):
        if d.factory:
            return {key: d.factory() for key in d.keys} | i
        if d.value is not None:
            return {key: d.value_copier(d.value) for key in d.keys} | i
        raise ValueError("No default value or factory specified.")
    return i


def compose_contexts(*contexts: Context) -> Context:
    """Compose contexts."""
    context = Context()
    for ctx in contexts:
        context |= ctx
    return context


def compose_context(*context_values: ContextValue) -> Context:
    """Compose inner contexts."""
    return Context({
        context_value.name_to_snake(): context_value for context_value in context_values
    })


def make_pipelines(
    typ: type,
    /,
    before: Sequence[AnyPipe] | AnyPipe | None = None,
    after: Sequence[AnyPipe] | AnyPipe | None = None,
) -> Context:
    """Compose defaults."""
    return compose_context(
        PipelineContext({typ: Pipelines.make(before=before, after=after)})
    )


def get_context_value(value_type: type[CV], context: Context | None) -> CV | None:
    """Get context values for a type."""
    return (context or Context()).get(get_context_key(value_type))  # pyright: ignore[reportReturnType]


def get_context_key(value_type: type[CV]) -> str:
    """Get context keys for a type."""
    return value_type.name_to_snake()



"""Pipes."""

from collections import UserDict
from collections.abc import Callable, Iterable, Mapping
from dataclasses import asdict, dataclass
from functools import wraps
from re import Pattern, compile
from string import whitespace
from typing import Any

from sympy import Symbol, symbols

from boilercv.correlations.types import CVL, Kind, P, Ps, R, SympifyParams, Transform
from boilercv.mappings import Repl, replace, replace_pattern, sort_by_keys_pattern
from boilercv.morphs.contexts import (
    Context,
    ContextValue,
    Defaults,
    Pipe,
    compose_context,
)
from boilercv.morphs.morphs import Morph


def fold_whitespace(
    forms: dict[Kind, str], defaults: Defaults[Kind, str]
) -> Morph[Kind, str]:
    """Fold whitespace."""
    return (
        Morph[Kind, str](forms)
        .morph_pipe(
            replace,
            (
                Repl[Kind](src=key, dst=key, find=find, repl=" ")
                for find in whitespace
                for key in defaults.keys
            ),
        )
        .morph_pipe(
            replace_pattern,
            (
                Repl[Kind](src=key, dst=key, find=r"\s+", repl=r" ")
                for key in defaults.keys
            ),
        )
        .morph_pipe(lambda forms: {k: v.strip() for k, v in forms.items()})
    )


class LocalSymbols(UserDict[str, Symbol], ContextValue):
    """Local symbols."""

    @classmethod
    def from_iterable(cls, syms: Iterable[str]):
        """Create from an iterable of symbols."""
        return cls(
            zip(
                syms,
                symbols(syms, nonnegative=True, real=True, finite=True),
                strict=True,
            )
        )


def set_latex_forms(forms: Morph[Kind, str]) -> Morph[Kind, str]:
    """Set forms for parameters."""
    if forms["sympy"] and not forms["latex"]:
        forms["latex"] = forms["sympy"]
    forms = forms.morph_pipe(
        replace,
        repls=[
            Repl[Kind](src="latex", dst="sympy", find=k, repl=v)
            for k, v in {r"_\b0": "_bo", r"_\o": "_0", "\\": ""}.items()
        ],
    )
    return forms


def compose_sympify_context(symbols: Iterable[str]) -> Context:
    """Get `sympify` parameters."""
    return compose_context(
        SympifyParams(locals=dict(LocalSymbols.from_iterable(symbols)), evaluate=False)
    )


@dataclass
class KeysPattern(ContextValue):
    """Keys pattern."""

    pattern: Pattern[str]
    groups: Iterable[str]
    apply_to_match: Callable[[list[str]], Any] = str
    message: str = "Match not found when sorting."


def contextualized(context_value_type: type[CVL]):
    """Contextualize function on `context_value_type`."""

    def contextualizer(f: Transform[P, Ps, R]) -> Callable[[P, CVL], R]:
        @wraps(f)
        def unpack_kwds(v: P, context_value: CVL) -> R:
            if not isinstance(context_value, context_value_type):
                raise ValueError(  # noqa: TRY004 so Pydantic catches it
                    f"Expected context value type '{context_value_type}', got '{type(context_value)}."
                )
            context_mapping = (
                context_value
                if isinstance(context_value, Mapping)
                else asdict(context_value)
            )
            return f(v, **context_mapping)  # pyright: ignore[reportCallIssue]

        return unpack_kwds

    return contextualizer


equation_name_pattern = compile(
    r"""(?xm)^  # x: verbose, m: multiline, ^: begin, always use multiline https://web.archive.org/web/20240429145610/https://sethmlarson.dev/regex-$-matches-end-of-string-or-newline
    (?P<author>[\w_]+?)  # lazy so year is leftmost match
    _(?P<sort>  # sort on year and optional equation number
        (?P<year>\d{4})  # year must be four digits
        (?:_(?P<num>[\d_]+))?  # optionally, multiple equations from one paper
    )
    $"""
)


def apply_to_match(groups: list[str]) -> tuple[int, int, str]:
    """Apply to match."""
    first, *rest = groups[0].split("_")
    year = int(first)
    num = int(rest[0]) if rest else 0
    author = groups[1]
    return year, num, author


sort_by_year = Pipe(
    contextualized(KeysPattern)(sort_by_keys_pattern),
    KeysPattern(
        pattern=equation_name_pattern,
        groups=["sort", "author"],
        apply_to_match=apply_to_match,
        message="Couldn't find year in equation identifier '{key}' when sorting.",
    ),
)



"""Image acquisition and processing."""

# * Pure numpy image processing functions take lots of types, including DataArrays.
# pyright: reportGeneralTypeIssues=none

from typing import Any

from matplotlib.font_manager import FontProperties, findfont
from numpy import asarray, iinfo, invert, mean, uint8
from numpy.typing import DTypeLike
from PIL import Image, ImageDraw, ImageFont, ImageOps

from boilercv.colors import BLACK, BLACK3, RED, WHITE, WHITE3
from boilercv.types import DA_T, Img, ImgLike

# * -------------------------------------------------------------------------------- * #
# * PURE NUMPY - TYPE PRESERVING


def scale_float(img: DA_T, dtype: DTypeLike = uint8) -> DA_T:
    """Return the input as `dtype` multiplied by the max value of `dtype`.

    Useful for scaling float-valued arrays to integer-valued images.
    """
    scaled = (img - img.min()) / (img.max() - img.min())
    return scaled.astype(dtype) * iinfo(dtype).max


def unpad(img: Img, pad_width: int) -> Img:
    """Remove padding from an image."""
    return img[pad_width:-pad_width, pad_width:-pad_width]


# * -------------------------------------------------------------------------------- * #
# * PURE NUMPY - NOT ALWAYS TYPE PRESERVING


def scale_bool(img: Any, dtype: DTypeLike = uint8) -> Any:
    """Return the input as `dtype` multiplied by the max value of `dtype`.

    Useful for functions (such as in OpenCV) which expect numeric bools.
    """
    return img.astype(dtype) * iinfo(dtype).max


# * -------------------------------------------------------------------------------- * #
# * OTHER - NOT ALWAYS TYPE PRESERVING

FONT = ImageFont.truetype(findfont(FontProperties(family="dejavu sans")), 24)
PAD = 10


def draw_text(image: Img, text: str = "") -> ImgLike:
    """Draw text in the top-right corner of an image.

    Args:
        image: Image.
        text: Text to draw.
    """
    if image.ndim == 3:
        rectangle_fill = BLACK3
        font_fill = WHITE3
    else:
        rectangle_fill = BLACK
        font_fill = WHITE
    _, image_width = image.shape[:2]
    pil_image = Image.fromarray(image)
    _, _, font_bbox_width, font_bbox_height = FONT.getbbox(text)
    text_p0 = (image_width - PAD - font_bbox_width, PAD)
    p0 = (text_p0[0] - PAD, text_p0[1] - PAD)
    p1 = (text_p0[0] + PAD + font_bbox_width, text_p0[1] + PAD + font_bbox_height)
    draw = ImageDraw.Draw(pil_image)
    draw.rectangle((p0, p1), fill=rectangle_fill)  # type: ignore  # pyright 1.1.348, pillow 10.2.0
    draw.text(text_p0, text, font=FONT, fill=font_fill)  # type: ignore  # pyright 1.1.348, pillow 10.2.0
    return asarray(pil_image)


def overlay(
    image: ImgLike, overlay: Img, color: tuple[int, int, int] = RED, alpha: float = 0.3
) -> Img:
    """Color an image given an overlay.

    Args:
        image: Image.
        overlay: Overlay image.
        color: Color for the overlay.
        alpha: Alpha value for the overlay. Range: 0-1
    """
    background = Image.fromarray(image).convert("RGBA")  # pyright: ignore[reportArgumentType] 1.1.356, pillow 10.0.0
    objects = Image.fromarray(overlay)
    if overlay.ndim == 2:
        objects = ImageOps.colorize(objects, WHITE3, color)  # type: ignore  # pyright 1.1.348, pillow 10.2.0
        mask = Image.fromarray(~(overlay * alpha).astype(uint8))
    else:
        avg: Img = mean(overlay, axis=-1)
        mask = Image.fromarray(invert(avg.astype(bool) * alpha).astype(uint8))
    composite = Image.composite(background, objects, mask)
    return asarray(composite)



"""Bubble collapse correlation models."""

from collections.abc import Callable, Iterable
from dataclasses import dataclass
from functools import partial
from typing import Any, Generic, get_args

import sympy
from numpy import nan
from pydantic import BaseModel, Field
from sympy.logic.boolalg import Boolean

from boilercv.correlations.pipes import (
    compose_sympify_context,
    fold_whitespace,
    set_latex_forms,
    sort_by_year,
)
from boilercv.correlations.types import (
    EQ,
    AnyEquation,
    Eq,
    Equation,
    Expectation,
    Expr,
    K,
    Kind,
    S,
)
from boilercv.morphs.contexts import (
    Context,
    ContextBaseModel,
    ContextMorph,
    Defaults,
    Pipe,
    compose_contexts,
    make_pipelines,
)
from boilercv.morphs.morphs import Morph


class EquationMetadata(BaseModel):
    """Meta."""

    name: str = ""
    """Name."""
    citekey: str = ""
    """Citekey."""


class Metadata(ContextMorph[AnyEquation, EquationMetadata]):
    """Equations."""

    @classmethod
    def get_context(cls) -> Context:
        """Get context."""
        return compose_contexts(
            cls.compose_defaults(), make_pipelines(cls, before=sort_by_year)
        )


class Expectations(ContextMorph[K, Expectation], Generic[K]):
    """Expectations."""

    @classmethod
    def get_context(cls) -> Context:
        """Get Pydantic context."""
        return cls.compose_defaults(value=nan)


class Forms(ContextMorph[Kind, str]):
    """Forms."""

    @classmethod
    def get_context(cls) -> Context:
        """Get context."""
        return compose_contexts(cls.compose_defaults(value=""))


class Params(ContextMorph[K, Forms], Generic[K]):
    """Parameters."""

    @classmethod
    def get_context(cls) -> Context:
        """Get Pydantic context."""
        return compose_contexts(
            cls.compose_defaults(value_context=Forms.get_context()),
            make_pipelines(Forms, after=[set_latex_forms]),
        )


def prep_equation_forms(
    i: dict[Kind, str], context: Context | None = None
) -> Morph[Kind, str]:
    """Prepare equation forms."""
    return Forms.context_model_validate(
        obj=i,
        context=(
            context
            or make_pipelines(
                Forms,
                before=[
                    Forms.get_context().pipelines[Forms].before[0],
                    Pipe(fold_whitespace, Defaults(keys=get_args(Kind))),
                ],
            )
        ),
    )


class EquationForms(ContextBaseModel, Generic[EQ]):
    """Equation forms."""

    latex: str
    """LaTeX form."""
    sympy: EQ
    """Symbolic form."""

    @classmethod
    def get_context(
        cls, symbols: Iterable[str] | None = None, forms_context: Context | None = None
    ) -> Context:
        """Get context."""
        symbols = symbols or ()
        return compose_contexts(
            make_pipelines(
                cls, before=partial(prep_equation_forms, context=forms_context)
            ),
            compose_sympify_context(symbols),  # If `EQ` is `Eq`
        )


class Equations(ContextMorph[Equation, EquationForms[EQ]], Generic[EQ]):
    """Equations."""

    @classmethod
    def get_context(
        cls, symbols: Iterable[str] | None = None, forms_context: Context | None = None
    ) -> Context:
        """Get context."""
        symbols = symbols or ()
        Eq_, *_ = cls.__pydantic_generic_metadata__["args"]  # noqa: N806
        return compose_contexts(
            cls.compose_defaults(
                value_context=EquationForms[Eq_].get_context(
                    symbols=symbols, forms_context=forms_context
                )
            ),
            make_pipelines(cls, before=sort_by_year),
        )


class Solutions(ContextBaseModel):
    """Solutions for a symbol."""

    solutions: list[Expr] = Field(default_factory=list)
    """Solutions."""
    warnings: list[str] = Field(default_factory=list)
    """Warnings."""

    @classmethod
    def get_context(cls, symbols: Iterable[str]) -> Context:
        """Get Pydantic context."""
        return compose_sympify_context(symbols)  # for `Expr`


class SymbolSolutions(ContextMorph[S, Solutions], Generic[S]):
    """Solutions for given symbols."""

    @classmethod
    def get_context(
        cls, symbols: Iterable[str], solve_syms: tuple[str, ...] | None = None
    ) -> Context:
        """Get Pydantic context."""
        return cls.compose_defaults(
            keys=solve_syms, value_context=Solutions.get_context(symbols=symbols)
        )


class EquationSolutions(ContextMorph[Equation, SymbolSolutions[S]], Generic[S]):
    """Equation solutions."""

    @classmethod
    def get_context(
        cls, symbols: Iterable[str], solve_syms: tuple[str, ...] | None = None
    ) -> Context:
        """Get Pydantic context."""
        SolveSyms, *_ = cls.__pydantic_generic_metadata__["args"]  # noqa: N806
        return compose_contexts(
            cls.compose_defaults(
                value_context=SymbolSolutions[SolveSyms].get_context(
                    symbols=symbols, solve_syms=solve_syms
                )
            ),
            make_pipelines(cls, before=sort_by_year),
        )


class SolvedEquations(ContextBaseModel, Generic[S]):
    """Solved equations."""

    equations: Equations[Eq]
    """Equations."""
    solutions: EquationSolutions[S]
    """Solutions."""

    @classmethod
    def get_context(
        cls, symbols: Iterable[str], solve_syms: tuple[str, ...]
    ) -> Context:
        """Get Pydantic context."""
        SolveSyms, *_ = cls.__pydantic_generic_metadata__["args"]  # noqa: N806
        return compose_contexts(
            Equations[Eq].get_context(symbols=symbols),
            EquationSolutions[SolveSyms].get_context(
                symbols=symbols, solve_syms=solve_syms
            ),
        )


@dataclass
class SymbolicCorrelation:
    """Correlation."""

    expr: sympy.Expr
    range: Boolean | None = None


@dataclass
class Correlation:
    """Correlation."""

    expr: Callable[..., Any]
    range: Callable[..., Any] | None



"""Models for image processing."""

from dataclasses import dataclass
from typing import Any

from boilercv.types import DA, SupportsMul


@dataclass
class Dimension:
    """A dimension.

    Args:
        dim: Dimension.
        long_name: Long name.
        coords: Coordinate values.
        units: Units.
        original_units: Original units to be converted from.
        original_coords: Original coordinate values to be converted from.
        scale: Scale factor to multiply the original coordinates by.
        parent_dim: Existing dimension to associate this dimension with.
    """

    dim: str
    """Dimension."""
    long_name: str = ""
    """Long name."""
    coords: Any = None
    """Coordinate values."""
    units: str = ""
    """Units."""
    original_units: str = ""
    """Original units to be converted from."""
    original_coords: SupportsMul | None = None
    """Original coordinate values to be converted from."""
    scale: float = 1
    """Scale factor to multiply the original coordinates by."""
    parent_dim: str = ""
    """Existing dimension to associate this dimension with."""

    @property
    def coordinate_units_match(self):
        """Whether the coordinate units are in the desired units."""
        return self.original_units == self.units

    def __post_init__(self):
        """Assign original units and long name if not specified."""
        if self.units and not self.original_units:
            self.original_units = self.units
        if not self.long_name:
            self.long_name = self.dim.capitalize() if len(self.dim) > 1 else self.dim
        if not self.parent_dim:
            self.parent_dim = self.dim

    def assign_to(self, da: DA) -> DA:
        """Assign this dimension and its coordinates."""
        if self.coords is None and self.dim == self.parent_dim:
            self.coords = da[self.parent_dim].values
        if not self.coordinate_units_match:
            self.convert()
        if self.coords is not None:
            da = da.assign_coords({self.dim: (self.parent_dim, self.coords)})
        attrs = {"long_name": self.long_name}
        if self.units:
            attrs["units"] = self.units
        da[self.dim] = da[self.dim].assign_attrs(attrs)
        return da

    def convert(self):
        """Convert the original coordinates to the desired units."""
        if self.original_coords is not None:
            self.coords = self.original_coords * self.scale
        self.original_units = self.units
        self.original_coords = self.coords


def get_dims(*dims: Dimension) -> tuple[str, ...]:
    """Get a tuple of dimension names representing these dimensions."""
    return tuple(dim.dim for dim in dims)



"""Dimensionless bubble diameter correlations for subcooled boiling bubble collapse."""

from pathlib import Path

from numpy import linspace, pi, sqrt

from boilercv import correlations
from boilercv.correlations.beta.types import SolveSym
from boilercv.correlations.models import Correlation, Expectations, SymbolicCorrelation
from boilercv.correlations.types import Equation, Sym

PNGS = Path("data/png_equations_beta")
"""Equation PNGs."""

_base = Path(__file__).with_suffix(".toml")

EQUATIONS_TOML = _base.with_stem("equations")
"""TOML file with equations."""
EXPECTATIONS_TOML = _base.with_stem("expectations")
"""TOML file with expectations."""
SOLUTIONS_TOML = _base.with_stem("solutions")
"""TOML file with solutions."""
SYMBOL_EXPECTATIONS = Expectations[Sym].context_model_validate(
    obj={
        "Nu_c": 1.0,
        "Fo_0": linspace(start=0.0, stop=5.0e-3, num=10),
        "Ja": 1.0,
        "Re_b": 10.0,
        "Re_b0": 100.0,
        "Pe": 1.0,
        "Pr": 1.0,
        "beta": 0.5,
        "alpha": 1.0,
        "pi": pi,
    },
    context=Expectations[Sym].get_context(),
)
"""Common keyword arguments applied to correlations.

A single test condition has been chosen to exercise each correlation across as wide of a
range as possible without returning `np.nan` values. This is done as follows:

- Let `Re_b0`, `Pr`, and `Ja` be 100.0, 1.0, and 1.0, respectively.
- Apply the correlation {eq}`eq_beta_tang_et_al_2016` with `Fo_0` such that `beta` is
  very close to zero. This is the correlation with the most rapidly vanishing value of
  `beta`.
- Choose ten linearly-spaced points for `Fo_0` between `0` and the maximum
  `Fo_0` just found.
"""


def get_equations_and_solutions() -> dict[Equation, SymbolicCorrelation]:
    """Get correlations."""
    return correlations.get_equations_and_solutions(
        EQUATIONS_TOML, SOLUTIONS_TOML, SolveSym
    )


def get_correlations() -> dict[Equation, Correlation]:
    """Get correlations."""
    return correlations.get_correlations(EQUATIONS_TOML, SOLUTIONS_TOML, SolveSym)


def florschuetz_chao_1965(bubble_fourier, bubble_jakob):
    """Florschuetz and Chao (1965) dimensionless bubble diameter {cite}`florschuetzMechanicsVaporBubble1965,tangReviewDirectContact2022`.

    Parameters
    ----------
    bubble_fourier
        Bubble Fourier number.
    bubble_jakob
        Bubble Jakob number.
    """
    return 1 - 4 * bubble_jakob * sqrt(bubble_fourier / pi)


def isenberg_sideman_1970(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - (3 / sqrt(pi))
        * bubble_initial_reynolds ** (1 / 2)
        * liquid_prandtl ** (1 / 3)
        * bubble_jakob
        * bubble_fourier
    ) ** (2 / 3)


def akiyama_1973(bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 1.036
        * bubble_fourier
        * bubble_initial_reynolds ** (1 / 2)
        * bubble_jakob
        * liquid_prandtl ** (1 / 3)
    ) ** 0.714


def chen_mayinger_1992(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 0.56
        * bubble_initial_reynolds**0.7
        * liquid_prandtl**0.5
        * bubble_jakob
        * bubble_fourier
    ) ** 0.9


def kalman_mori_2002(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 0.0094
        * bubble_initial_reynolds**0.855
        * liquid_prandtl**0.855
        * bubble_jakob
        * bubble_fourier
    ) ** 0.873


def yuan_et_al_2009(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`yuandewenCondensationHeatTransfer2009,tangReviewDirectContact2022`."""
    return (
        1
        - 1.8
        * bubble_initial_reynolds**0.5
        * liquid_prandtl ** (1 / 3)
        * bubble_jakob
        * bubble_fourier
        * (1 - 0.5 * bubble_jakob**0.1 * bubble_fourier)
    ) ** (2 / 3)


def lucic_mayinger_2010(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 2.92
        * bubble_initial_reynolds**0.61
        * liquid_prandtl**0.33
        * bubble_jakob**0.69
        * bubble_fourier
    )


def kim_park_2011(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 0.67
        * bubble_initial_reynolds**0.7
        * liquid_prandtl ** (-0.4564)
        * bubble_jakob**0.7959
        * bubble_fourier
    ) ** 0.769


# def inaba_et_al_2013(
#     bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
# ):
#     """Bubble history correlation for condensation of a stagnant bubble. {cite}`tangReviewDirectContact2022`."""
#     return (
#         1
#         - 1.1
#         * bubble_initial_reynolds**0.86
#         * liquid_prandtl ** (2 / 3)
#         * bubble_jakob**0.2
#         * bubble_fourier
#     )


def al_issa_et_al_2014(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 0.135
        * bubble_initial_reynolds**0.89
        * liquid_prandtl**0.33
        * bubble_jakob
        * bubble_fourier
    ) ** 0.901


def tang_et_al_2016(
    bubble_fourier, bubble_initial_reynolds, liquid_prandtl, bubble_jakob
):
    """Bubble history correlation for condensation of a stagnant bubble {cite}`tangReviewDirectContact2022`."""
    return (
        1
        - 12.29
        * bubble_initial_reynolds**0.584
        * liquid_prandtl**0.333
        * bubble_jakob**0.581
        * bubble_fourier
    ) ** 0.706



"""Types used in {mod}`~boilercv.morphs`."""

from collections.abc import Callable, MutableMapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    NamedTuple,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    _LiteralGenericAlias,  # pyright: ignore[reportAttributeAccessIssue]
    _UnionGenericAlias,  # pyright: ignore[reportAttributeAccessIssue]
    runtime_checkable,
)

from pydantic import BaseModel

if TYPE_CHECKING:
    from boilercv.morphs.contexts import (
        ContextValue,
        Pipe,
        PipelineContext,
        PipeWithInfo,
    )

AnyPipe: TypeAlias = "Pipe | PipeWithInfo | Callable[[Any], Any]"
"""Any pipe."""
Mode: TypeAlias = Literal["before", "after"]
"""Mode."""
UnionGenericAlias: TypeAlias = _UnionGenericAlias
"""Union type."""
LiteralGenericAlias: TypeAlias = _LiteralGenericAlias
"""Literal type."""
ContextValueLike: TypeAlias = "PipelineContext | ContextValue"
"""Allowable context-value-like values."""

T = TypeVar("T")
"""Type."""
P = TypeVar("P", contravariant=True)
"""Contravariant type to represent parameters."""
R = TypeVar("R", covariant=True)
"""Covariant type to represent returns."""
Ps = ParamSpec("Ps")
"""Parameter type specification."""
K = TypeVar("K")
"""Key type."""
V = TypeVar("V")
"""Value type."""
RK = TypeVar("RK")
"""Return key type that may differ from parameter key type."""
RV = TypeVar("RV")
"""Return value type that may differ from parameter value type."""
CV = TypeVar("CV", bound="ContextValue", contravariant=True)
"""Context value type."""
Model = TypeVar("Model", bound=BaseModel)
"""Model type."""
TypeOfModel = TypeVar("TypeOfModel", bound=type[BaseModel])
"""Type of model type."""


class TypeType(Protocol[P, R, Ps]):  # noqa: D101
    def __call__(self, i: P, /, *args: Ps.args, **kwds: Ps.kwargs) -> R: ...  # noqa: D102


class TypeMap(Protocol[P, K, V, Ps]):  # noqa: D101
    def __call__(  # noqa: D102
        self, i: P, /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> MutableMapping[K, V]: ...


class TypeDict(Protocol[P, K, V, Ps]):  # noqa: D101
    def __call__(  # noqa: D102
        self, i: P, /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> dict[K, V]: ...


class MapType(Protocol[K, V, R, Ps]):  # noqa: D101
    def __call__(  # noqa: D102
        self, i: MutableMapping[K, V], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> R: ...


class DictType(Protocol[K, V, R, Ps]):  # noqa: D101
    def __call__(  # noqa: D102
        self, i: dict[K, V], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> R: ...


class MapMap(Protocol[K, V, RK, RV, Ps]):  # noqa: D101
    def __call__(  # noqa: D102
        self, i: MutableMapping[K, V], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> MutableMapping[RK, RV]: ...


class DictDict(Protocol[K, V, RK, RV, Ps]):  # noqa: D101
    def __call__(  # noqa: D102
        self, i: dict[K, V], /, *args: Ps.args, **kwds: Ps.kwargs
    ) -> dict[RK, RV]: ...


class Types(NamedTuple):
    """Mapping types."""

    key: type
    value: type


@runtime_checkable
class HasModelDump(Protocol):
    """Has model dump."""

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",  # noqa: PYI051
        include: Any = None,
        exclude: Any = None,
        context: dict[str, Any] | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        serialize_as_any: bool = False,
    ) -> Any:
        """Model dump."""



"""Packing and unpacking of binarized video data."""

from numpy import packbits, unpackbits
from xarray import apply_ufunc

from boilercv.data import (
    DIMS,
    PACKED,
    PACKED_DIM_INDEX,
    PACKED_DIMS,
    VIDEO,
    XPX,
    XPX_PACKED,
)
from boilercv.types import DA


def pack(da: DA) -> DA:
    """Pack the bits in dimension of the data array."""
    return (
        apply_ufunc(
            packbits,
            da,
            kwargs=dict(axis=PACKED_DIM_INDEX),
            input_core_dims=[DIMS],
            output_core_dims=[DIMS],
            exclude_dims={XPX},
            keep_attrs=True,
        )
        .rename({XPX: XPX_PACKED})
        .rename(f"{VIDEO}_{PACKED}")
    )


def unpack(da: DA) -> DA:
    """Unpack the bits of the last image dimension of a data array."""
    return (
        apply_ufunc(
            unpackbits,
            da,
            kwargs=dict(axis=PACKED_DIM_INDEX),
            input_core_dims=[PACKED_DIMS],
            output_core_dims=[PACKED_DIMS],
            exclude_dims={XPX_PACKED},
            keep_attrs=True,
        )
        .rename({XPX_PACKED: XPX})
        .rename(VIDEO)
        .astype(bool)
    )
